# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals


class MetaSingleCellOperandCache(type):
    def __call__(cls, row, column, row_fixed=False, column_fixed=False, ws_name=None, source=None):
        if source is not None and ws_name is not None and source._caches:
            cache = source._caches['single']
            key = (ws_name, row, column, row_fixed, column_fixed)

            try:
                return cache[key]
            except KeyError:
                value = cache[key] = super(MetaSingleCellOperandCache, cls).__call__(row, column, row_fixed=row_fixed,
                                                                                     column_fixed=column_fixed,
                                                                                     ws_name=ws_name,
                                                                                     source=source)
                return value
        else:
            return super(MetaSingleCellOperandCache, cls).__call__(row, column, row_fixed=row_fixed,
                                                                   column_fixed=column_fixed, ws_name=ws_name,
                                                                   source=source)


class MetaCellRangeOperandCache(type):
    def __call__(cls, row1, column1, row2, column2,
                 row1_fixed=False, column1_fixed=False, row2_fixed=False, column2_fixed=False,
                 ws_name=None, source=None):
        if source is not None and ws_name is not None and source._caches:
            cache = source._caches['range']
            key = (ws_name, row1, column1, row2, column2, row1_fixed, column1_fixed, row2_fixed, column2_fixed)

            try:
                return cache[key]
            except KeyError:
                value = cache[key] = super(MetaCellRangeOperandCache, cls).__call__(
                    row1, column1, row2, column2, row1_fixed, column1_fixed, row2_fixed, column2_fixed, ws_name, source
                )
                return value
        else:
            return super(MetaCellRangeOperandCache, cls).__call__(row1, column1, row2, column2, row1_fixed,
                                                                  column1_fixed, row2_fixed, column2_fixed, ws_name,
                                                                  source)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from abc import ABCMeta, abstractmethod

from six import add_metaclass

from efc import Lexer, Parser
from efc.interfaces.cache import CacheManager
from efc.interfaces.errors import NamedRangeNotFound


@add_metaclass(ABCMeta)
class BaseExcelInterface(object):
    """
    Base class to working with excel document
    """

    def __init__(self, use_cache=False, lexer=Lexer, parser=Parser):
        self._cache_manager = CacheManager() if use_cache else None

        self.lexer = lexer()
        self.parser = parser()

    def _build_rpn(self, formula, ws_name=None):
        tokens_line = self.lexer.parse(formula)
        return self.parser.to_rpn(tokens_line, ws_name=ws_name, source=self)

    def _calc_formula(self, formula, ws_name=None):
        """
        Calculate formula
        :type formula: str
        :type ws_name: str
        """
        rpn = self._build_rpn(formula, ws_name)
        return rpn.calc(ws_name, self)

    @property
    def _caches(self):
        return self._cache_manager

    def clear_cache(self):
        """Clear all caches"""
        self._caches.clear()

    @abstractmethod
    def _cell_to_value(self, row, column, ws_name):
        """
        :type row: int
        :type column: int
        :type ws_name: basestring
        :rtype: list
        """

    @abstractmethod
    def _get_named_range_formula(self, name, ws_name):
        """
        Should raise NamedRangeNotFound if named range not found
        :type name: basestring
        :type ws_name: basestring
        :rtype: basestring
        """

    @abstractmethod
    def _max_row(self, ws_name):
        pass

    @abstractmethod
    def _min_row(self, ws_name):
        pass

    @abstractmethod
    def _max_column(self, ws_name):
        pass

    @abstractmethod
    def _min_column(self, ws_name):
        pass

    @abstractmethod
    def _has_worksheet(self, ws_name):
        pass

    def _named_range_to_cells(self, name, ws_name):
        f = self._get_named_range_formula(name, ws_name)
        return self._calc_formula(f, ws_name)

    def _has_named_range(self, name, ws_name):
        try:
            self._get_named_range_formula(name, ws_name)
        except NamedRangeNotFound:
            return False
        else:
            return True



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import re
from calendar import isleap, monthrange
from functools import wraps

from six import integer_types, string_types
from six.moves import range, zip_longest

from efc.rpn_builder.parser.operands import (
    BadReference, CellRangeOperand, CellSetOperand, EmptyOperand, ErrorOperand, NamedRangeOperand, NotFoundErrorOperand,
    NumErrorOperand, RPNOperand, SetOperand, SimpleOperand, SimpleSetOperand, SingleCellOperand, ValueErrorOperand,
    ValueNotAvailable,
)
from efc.utils import is_float, parse_date

__all__ = ('EXCEL_FUNCTIONS',)


def _get_type_id(obj):
    if isinstance(obj, bool):
        return 2
    elif isinstance(obj, string_types):
        return 1
    return 0


def type_mixin(a, b):
    if a is None:
        a_type = (1, '') if isinstance(b, string_types) else (0, 0)
    else:
        a_type = (_get_type_id(a), a)

    if b is None:
        b_type = (1, '') if isinstance(a, string_types) else (0, 0)
    else:
        b_type = (_get_type_id(b), b)
    return a_type, b_type


def set_mixin(foo):
    @wraps(foo)
    def wrapper(op1, op2):
        fill_value = None

        op1_is_set = isinstance(op1, SetOperand)
        op2_is_set = isinstance(op2, SetOperand)

        if op1_is_set or op2_is_set:
            if not op1_is_set:
                op1, fill_value = tuple(), op1
            elif not op2_is_set:
                op2, fill_value = tuple(), op2

            result = SimpleSetOperand()
            for v1, v2 in zip_longest(op1, op2, fillvalue=fill_value):
                result.add_cell(ValueNotAvailable() if v1 is None or v2 is None else SimpleOperand(foo(v1, v2)))
        else:
            result = foo(op1, op2)
        return result

    return wrapper


@set_mixin
def add(op1, op2):
    return op1.digit + op2.digit


def add_func(op1, op2=None):
    if isinstance(op2, EmptyOperand):
        op2 = None

    if op2 is not None:
        return add(op1, op2)
    else:
        return op1.digit


@set_mixin
def sub(op1, op2):
    return op1.digit - op2.digit


def subtract_func(op1, op2=None):
    if isinstance(op2, EmptyOperand):
        op2 = None

    if op2 is not None:
        return sub(op1, op2)
    else:
        return -op1.digit


@set_mixin
def divide_func(op1, op2):
    return op1.digit / op2.digit


@set_mixin
def multiply_func(op1, op2):
    return op1.digit * op2.digit


@set_mixin
def concat_func(op1, op2):
    if isinstance(op1, ErrorOperand):
        raise op1
    if isinstance(op2, ErrorOperand):
        raise op2
    return op1.string + op2.string


@set_mixin
def exponent_func(op1, op2):
    return op1.digit ** op2.digit


@set_mixin
def compare_not_eq_func(op1, op2):
    op1, op2 = type_mixin(op1.value, op2.value)
    return op1 != op2


@set_mixin
def compare_gte_func(op1, op2):
    op1, op2 = type_mixin(op1.value, op2.value)
    return op1 >= op2


@set_mixin
def compare_lte_func(op1, op2):
    op1, op2 = type_mixin(op1.value, op2.value)
    return op1 <= op2


@set_mixin
def compare_gt_func(op1, op2):
    op1, op2 = type_mixin(op1.value, op2.value)
    return op1 > op2


@set_mixin
def compare_lt_func(op1, op2):
    op1, op2 = type_mixin(op1.value, op2.value)
    return op1 < op2


@set_mixin
def compare_eq_func(op1, op2):
    op1, op2 = type_mixin(op1.value, op2.value)
    return op1 == op2


def iter_elements(*args):
    for arg in args:
        if isinstance(arg, RPNOperand):
            arg = arg.evaluated_value

        if isinstance(arg, (CellRangeOperand, CellSetOperand)):
            for cell in arg:
                yield cell
        else:
            yield arg


def iter_digits(*args):
    for op in iter_elements(*args):
        if not op.is_blank:
            try:
                yield op.digit
            except (ValueError, TypeError):
                pass


def sum_func(*args):
    return sum(d or 0 for d in iter_digits(*args))


def mod_func(op1, op2):
    return op1.digit % op2.digit


def if_func(expr_op, op1, op2=None):
    if isinstance(op2, EmptyOperand):
        op2 = None

    if op2 is None:
        op2 = False
    return op1 if expr_op.value else op2


def ifs_func(*args):
    for i in range(0, len(args), 2):
        if args[i].value:
            return args[i + 1].value
    else:
        raise ValueNotAvailable


def if_error_func(op1, op2):
    if isinstance(op1, RPNOperand):
        op1 = op1.evaluated_value
    return op2 if isinstance(op1, ErrorOperand) else op1


def is_error_func(op):
    if isinstance(op, RPNOperand):
        op = op.evaluated_value
    return isinstance(op, ErrorOperand)


def max_func(*args):
    return max(list(d or 0 for d in iter_digits(*args)) or [0])


def min_func(*args):
    return min(list(d or 0 for d in iter_digits(*args)) or [0])


def left_func(op1, op2=1):
    if isinstance(op2, EmptyOperand):
        op2 = 1

    return op1.string[:int(op2)]


def right_func(op1, op2=1):
    if isinstance(op2, EmptyOperand):
        op2 = 1

    return op1.string[-int(op2):]


def mid_func(op1, op2, op3):
    left = int(op2) - 1
    right = left + int(op3)
    return op1.string[left:right]


def is_blank_func(a):
    while True:
        if isinstance(a, (NamedRangeOperand, SingleCellOperand)):
            a = SimpleOperand(a.address_to_value())
        elif isinstance(a, RPNOperand):
            a = a.evaluated_value
        else:
            break

    try:
        it = iter(a)
    except TypeError:
        return a.value is None

    new_set = SimpleSetOperand()
    for v in it:
        new_set.add_cell(SimpleOperand(v.value is None))
    return new_set


def or_function(*args):
    for op in iter_elements(*args):
        v = op.value
        if v is not None and not isinstance(v, string_types) and v:
            return True
    return False


def and_function(*args):
    for op in iter_elements(*args):
        v = op.value
        if v is not None and not isinstance(v, string_types) and not v:
            return False
    return True


def not_func(op):
    return not op.value


def small_function(r, op):
    items = sorted(iter_digits(r))
    index = int(op) - 1
    try:
        return items[index]
    except IndexError:
        return ValueErrorOperand()


def large_function(r, op):
    items = sorted(iter_digits(r), reverse=True)
    index = int(op) - 1
    try:
        return items[index]
    except IndexError:
        return ValueErrorOperand()


def round_function(a, b):
    b = int(b)
    v = round(a.digit, b)
    if b == 0:
        v = int(v)
    return v


def round_down_function(a, b):
    b = int(b)
    base = 10 ** b
    v = a.digit * base // 1 / base
    if b == 0:
        v = int(v)
    return v


def floor_function(a, multiple):
    multiple = int(multiple)
    return int(a.digit / multiple) * multiple


def count_function(*args):
    return len([op for op in iter_elements(*args) if isinstance(op.value, (integer_types, float))])


def abs_function(a):
    return abs(a.digit)


def match_function(op1, r, match_type=None):
    if isinstance(match_type, EmptyOperand):
        match_type = None

    match_type = 0 if match_type is None else int(match_type)

    expr = op1.value
    idx = None
    if match_type == 1:
        for idx, item in enumerate(r, 1):
            a, b = type_mixin(item.value, expr)
            if a == b:
                break
            elif a > b:
                idx -= 1
                break
    elif match_type == -1:
        for idx, item in enumerate(r, 1):
            a, b = type_mixin(item.value, expr)
            if a > b:
                break
        else:
            idx = None
    else:
        for idx, item in enumerate(r, 1):
            a, b = type_mixin(item.value, expr)
            if a == b:
                break
        else:
            idx = None
    if idx is None:
        raise NotFoundErrorOperand()
    return idx


COUNT_IF_EXPR = re.compile(r'^(?P<symbol><=|>=|<>|>|<|=)(?P<value>.+)$')


def get_check_function(expr):
    if isinstance(expr.value, string_types):
        match = COUNT_IF_EXPR.search(expr.value)
        if match:
            match = match.groupdict()
            operation = match['symbol']
            operand = match['value']
        else:
            operation = '='
            operand = expr.value
    else:
        operation = '='
        operand = expr.value

    if is_float(operand):
        operand = float(operand)

    check = ARITHMETIC_FUNCTIONS[operation]
    return check, SimpleOperand(operand)


def countif_function(cells, expr):
    check, operand = get_check_function(expr)
    return len([op for op in cells.value if check(op, operand)])


def counta_function(cells):
    return len([op for op in cells.value if op.value is not None])


def get_checks_from_args(args):
    args = iter(args)
    checks = []
    while True:
        check = []
        try:
            check.append(next(args))
        except StopIteration:
            break

        check.extend(get_check_function(next(args)))
        checks.append(check)
    return checks


def ifs_indexes(*args):
    good_indexes = None

    checks = get_checks_from_args(args)

    for op_range, check, expr in checks:
        check_good_indexes = None
        key = None
        if op_range.source and isinstance(op_range, CellRangeOperand):
            cache = op_range.source._caches['ifs'] if op_range.source._caches is not None else None

            if cache is not None:
                key = (op_range.ws_name, op_range.row1, op_range.column1, op_range.row2,
                       op_range.column2, check, expr.value)
                check_good_indexes = cache.get(key)
        else:
            cache = None

        if check_good_indexes is None:
            check_good_indexes = set()
            for idx, item in enumerate(op_range, 1):
                if item is None:
                    raise ValueErrorOperand
                else:
                    # convert expr value to item type
                    if isinstance(item.value, string_types) and not isinstance(expr.value, string_types):
                        expr = SimpleOperand(expr.string)
                    elif isinstance(item.value, (integer_types, float)) and not isinstance(expr.value,
                                                                                           (integer_types, float)):
                        try:
                            expr = SimpleOperand(expr.digit)
                        except ValueError:
                            pass

                    if check(item, expr):
                        check_good_indexes.add(idx)

            if cache is not None:
                cache[key] = check_good_indexes

        if good_indexes is None:
            good_indexes = check_good_indexes.copy()
        else:
            good_indexes &= check_good_indexes

        if not good_indexes:
            break

    return good_indexes


def sum_ifs_function(op1, *args):
    good_indexes = ifs_indexes(*args)
    return sum_func(*(c for idx, c in enumerate(op1, 1) if idx in good_indexes))


def sum_if_function(r, expr, op1):
    return sum_ifs_function(op1, r, expr)


def concatenate(*args):
    return ''.join(i.string for i in iter_elements(*args))


def average_function(*args):
    values = list(iter_digits(*args))
    return sum(values) / len(values)


def average_ifs_function(op1, *args):
    good_indexes = ifs_indexes(*args)
    return average_function(*(c for idx, c in enumerate(op1, 1) if idx in good_indexes))


def count_blank_function(cells):
    return len([op for op in iter_elements(cells) if op.is_blank])


def count_ifs_function(op1, *args):
    good_indexes = ifs_indexes(*args)
    return count_function(*(c for idx, c in enumerate(op1, 1) if idx in good_indexes))


def offset_function(cell, row_offset, col_offset, height=None, width=None):
    height_is_none = height is None or isinstance(height, EmptyOperand)
    width_is_none = width is None or isinstance(width, EmptyOperand)

    if isinstance(cell, SingleCellOperand):
        height = int(height) if not height_is_none else 1
        width = int(width) if not width_is_none else 1

        column = cell.column
        row = cell.row
    elif isinstance(cell, CellRangeOperand):
        height = int(height) if not height_is_none else cell.row2 - cell.row1 + 1
        width = int(width) if not width_is_none else cell.column2 - cell.column1 + 1

        column = cell.column1
        row = cell.row1
    else:
        return ValueErrorOperand()

    if height == width == 1:
        return SingleCellOperand(row=row + int(row_offset), column=column + int(col_offset),
                                 ws_name=cell.ws_name, source=cell.source)
    else:
        return CellRangeOperand(row1=row + int(row_offset),
                                column1=column + int(col_offset),
                                row2=row + int(row_offset) + height - 1,
                                column2=column + int(col_offset) + width - 1,
                                ws_name=cell.ws_name, source=cell.source)


def vlookup_function(op, rg, column, flag=None):
    if isinstance(flag, EmptyOperand):
        flag = None

    first_col = rg.offset()
    first_col.column2 = first_col.column1

    if flag is not None and flag.digit or flag is None:
        idx = match_function(op, first_col, 1)
        if flag is not None and flag.digit and idx != 1:
            a, b = type_mixin(list(first_col)[idx - 1].value, op.value)
            if a == b:
                idx -= 1
    else:
        idx = match_function(op, first_col, 0)
    return SingleCellOperand(row=(rg.row1 or 1) + idx - 1, column=(rg.column1 or 1) + column.digit - 1,
                             ws_name=rg.ws_name, source=rg.source)


def index_function(rg, row, column=None):
    if isinstance(column, EmptyOperand):
        column = None

    if isinstance(rg, RPNOperand):
        rg = rg.evaluated_value

    if isinstance(rg, SetOperand):
        set_type = SimpleSetOperand
        row1 = 1
        row2 = rg.rows_count
        column1 = 1
        column2 = rg.columns_count
    else:
        set_type = CellSetOperand
        row1 = rg.row1
        row2 = rg.row2
        column1 = rg.column1
        column2 = rg.column2

    rg_size = 2 if row1 != row2 and column1 != column2 else 1

    row = row.digit
    if column is not None:
        column = column.digit
    elif rg_size == 1:
        column = 1

    if rg_size == 1:
        if column1 is not None and column != 1:
            return BadReference()
    else:
        if column is None or row == 0 or column == 0:
            return BadReference()

    if row == 0:
        result = set_type()
        for c in range(column1, column2 + 1):
            result.add_cell(rg.get_cell(1, c))
    elif column == 0:
        result = set_type()
        for r in range(row1, row2 + 1):
            result.add_cell(rg.get_cell(r, 1))
    else:
        result = rg.get_cell(row, column)

    return result


def substitute_func(text, old_text, new_text, instance_num=None):
    if isinstance(instance_num, EmptyOperand):
        instance_num = None

    instance_num = instance_num.digit if instance_num is not None else -1
    return text.string.replace(old_text.string, new_text.string, instance_num)


def search_func(pattern, source, start_position=None):
    if isinstance(start_position, EmptyOperand):
        start_position = None

    try:
        start_position = start_position or 0
        return source.string[start_position:].lower().index(pattern.string.lower())
    except ValueError:
        return ValueErrorOperand()


TRIM_REGEXP = re.compile(r' {2,}')


def trim_func(op):
    value = op.string.strip()
    value = TRIM_REGEXP.sub(' ', value)
    return value


def len_func(op):
    value = op.string
    return len(value)


def year_frac(dt1, dt2, tp=None):
    if isinstance(tp, EmptyOperand) or tp is None:
        tp = 0
    else:
        tp = tp.digit

    if tp in (0, 4):
        dt1p = parse_date(dt1.string)
        dt2p = parse_date(dt2.string)

        dt1p_day = dt1p.day
        dt2p_day = dt2p.day

        # https://en.wikipedia.org/wiki/Day_count_convention#30/360_US
        if tp == 0:
            lfd1 = monthrange(dt1p.year, dt1p.month)[1]
            lfd2 = monthrange(dt2p.year, dt1p.month)[1]
            if dt1p_day == lfd1 and dt2p_day == lfd2:
                dt2p_day = 30

            if dt1p_day == lfd1:
                dt1p_day = 30

            if dt2p_day == 31 and dt1p_day >= 30:
                dt2p_day = 30

            if dt1p_day == 31:
                dt1p_day = 30

        # https://en.wikipedia.org/wiki/Day_count_convention#30E/360
        else:
            if dt1p_day == 31:
                dt1p_day = 30

            if dt2p_day == 31:
                dt2p_day = 30

        # https://en.wikipedia.org/wiki/Day_count_convention#30/360_methods
        v = (360 * (dt2p.year - dt1p.year) + 30 * (dt2p.month - dt1p.month) + (dt2p_day - dt1p_day)) / 360
    elif tp in (1, 2, 3):
        if tp == 1:
            year_days = 366 if isleap(parse_date(dt2.string).year) else 365
        elif tp == 2:
            year_days = 360
        else:
            year_days = 365
        v = (int(dt2.digit) - int(dt1.digit)) / year_days
    else:
        v = NumErrorOperand()

    return v


ARITHMETIC_FUNCTIONS = {
    '+': add_func,
    '-': subtract_func,
    '/': divide_func,
    '*': multiply_func,
    '&': concat_func,
    '^': exponent_func,
    '<>': compare_not_eq_func,
    '>=': compare_gte_func,
    '<=': compare_lte_func,
    '>': compare_gt_func,
    '<': compare_lt_func,
    '=': compare_eq_func,
}

EXCEL_FUNCTIONS = {}
EXCEL_FUNCTIONS.update(ARITHMETIC_FUNCTIONS)

EXCEL_FUNCTIONS['ABS'] = abs_function
EXCEL_FUNCTIONS['AND'] = and_function
EXCEL_FUNCTIONS['AVERAGE'] = average_function
EXCEL_FUNCTIONS['AVERAGEIFS'] = average_ifs_function

EXCEL_FUNCTIONS['CONCATENATE'] = concatenate
EXCEL_FUNCTIONS['COUNT'] = count_function
EXCEL_FUNCTIONS['COUNTA'] = counta_function
EXCEL_FUNCTIONS['COUNTIF'] = countif_function
EXCEL_FUNCTIONS['COUNTIFS'] = count_ifs_function
EXCEL_FUNCTIONS['COUNTBLANK'] = count_blank_function

EXCEL_FUNCTIONS['FLOOR'] = floor_function

EXCEL_FUNCTIONS['IF'] = if_func
EXCEL_FUNCTIONS['IFS'] = ifs_func
EXCEL_FUNCTIONS['IFERROR'] = if_error_func
EXCEL_FUNCTIONS['INDEX'] = index_function
EXCEL_FUNCTIONS['ISBLANK'] = is_blank_func
EXCEL_FUNCTIONS['ISERROR'] = is_error_func

EXCEL_FUNCTIONS['LARGE'] = large_function
EXCEL_FUNCTIONS['LEN'] = len_func
EXCEL_FUNCTIONS['LEFT'] = left_func

EXCEL_FUNCTIONS['MATCH'] = match_function
EXCEL_FUNCTIONS['MAX'] = max_func
EXCEL_FUNCTIONS['MID'] = mid_func
EXCEL_FUNCTIONS['MIN'] = min_func
EXCEL_FUNCTIONS['MOD'] = mod_func

EXCEL_FUNCTIONS['NOT'] = not_func

EXCEL_FUNCTIONS['OFFSET'] = offset_function
EXCEL_FUNCTIONS['OR'] = or_function

EXCEL_FUNCTIONS['RIGHT'] = right_func
EXCEL_FUNCTIONS['ROUND'] = round_function
EXCEL_FUNCTIONS['ROUNDDOWN'] = round_down_function

EXCEL_FUNCTIONS['SEARCH'] = search_func
EXCEL_FUNCTIONS['SMALL'] = small_function
EXCEL_FUNCTIONS['SUBSTITUTE'] = substitute_func
EXCEL_FUNCTIONS['SUM'] = sum_func
EXCEL_FUNCTIONS['SUMIF'] = sum_if_function
EXCEL_FUNCTIONS['SUMIFS'] = sum_ifs_function

EXCEL_FUNCTIONS['TRIM'] = trim_func

EXCEL_FUNCTIONS['VLOOKUP'] = vlookup_function

EXCEL_FUNCTIONS['YEARFRAC'] = year_frac



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import re
from collections import OrderedDict

from efc.rpn_builder.lexer import tokens
from efc.rpn_builder.lexer.errors import CheckSumError
from efc.utils import Array

__all__ = ('Lexer',)

TOKENS_PRIORITY = (
    tokens.FloatToken,
    tokens.IntToken,
    tokens.BoolToken,
    tokens.StringToken,
    tokens.FunctionToken,
    tokens.CellsRangeToken,
    tokens.SingleCellToken,
    tokens.NamedRangeToken,
    tokens.AddToken,
    tokens.SubtractToken,
    tokens.DivideToken,
    tokens.MultiplyToken,
    tokens.ConcatToken,
    tokens.ExponentToken,
    tokens.CompareNotEqToken,
    tokens.CompareGTEToken,
    tokens.CompareLTEToken,
    tokens.CompareGTToken,
    tokens.CompareLTToken,
    tokens.CompareEqToken,
    tokens.LeftBracketToken,
    tokens.RightBracketToken,
    tokens.SpaceToken,
    tokens.Separator,
)


class TokensLine(Array):
    def __init__(self, line):
        self.src_line = line
        super(TokensLine, self).__init__()


class Lexer(object):
    def __init__(self):
        self.lexer_tokens = OrderedDict()
        self.regexp = None

        self.prepare_regexp()

    def prepare_regexp(self):
        regexp_list = []
        for c in TOKENS_PRIORITY:
            self.lexer_tokens[c.__name__] = c
            regexp_list.append(c.get_group_pattern())
        self.regexp = r'|'.join(regexp_list)

    def parse(self, line):
        lexer_tokens = self.lexer_tokens
        parsed_line = []

        tokens_line = TokensLine(line)
        for match in re.finditer(self.regexp, line, flags=re.UNICODE):
            token = lexer_tokens[match.lastgroup](match)
            if not isinstance(token, tokens.SpaceToken):
                tokens_line.append(token)
            parsed_line.append(token.src_value)

        parsed_line = ''.join(parsed_line)
        if parsed_line != line:
            raise CheckSumError(line, parsed_line)
        return tokens_line



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from efc.interfaces.base import BaseExcelInterface
from efc.interfaces.errors import NamedRangeNotFound
from efc.utils import datetime_to_openxml, parse_date
from openpyxl.utils.cell import coordinate_to_tuple


class OpenpyxlInterface(BaseExcelInterface):
    def __init__(self, wb, *args, **kwargs):
        self.wb = wb
        super(OpenpyxlInterface, self).__init__(*args, **kwargs)

    def _serialize_value(self, value, data_type):
        if data_type == 'd':
            value = datetime_to_openxml(value)
        return value

    def _deserialize_value(self, value, data_type):
        if data_type == 'd':
            value = parse_date(value)
        return value

    def calc_cell(self, cell_index, ws_name):
        """
        Calculate the cell formula by str index.
        If the cell does not have the formula it returns the cell's value.
        :type cell_index: basestring
        :type ws_name: basestring
        """
        row, column = coordinate_to_tuple(cell_index)
        value = self._cell_to_value(row, column, ws_name)
        cell = self.wb[ws_name]._get_cell(row, column)
        return self._deserialize_value(value, cell.data_type)

    def _cell_to_value(self, row, column, ws_name):
        cell = self.wb[ws_name]._get_cell(row, column)

        if cell.data_type != 'f':
            return self._serialize_value(cell.value, cell.data_type)
        elif not self._caches:
            f = cell.value[1:]
            return self._calc_formula(f, ws_name).value
        else:
            cache_key = (ws_name, row, column)
            if cache_key not in self._caches['cells']:
                f = cell.value[1:]
                self._caches['cells'][cache_key] = self._calc_formula(f, ws_name).value

            return self._caches['cells'][cache_key]

    def _get_named_range_formula(self, name, ws_name):
        local_sheet_id = None
        if ws_name is not None:
            local_sheet_id = self.wb.sheetnames.index(ws_name)

        result = None
        for named_range in self.wb.defined_names.definedName:
            if named_range.name == name:
                result = named_range.attr_text
                if named_range.localSheetId == local_sheet_id:
                    break

        if result is None:
            raise NamedRangeNotFound
        else:
            return result

    def _max_row(self, ws_name):
        return self.wb[ws_name]._max_row

    def _min_row(self, ws_name):
        return self.wb[ws_name]._min_row

    def _max_column(self, ws_name):
        return self.wb[ws_name]._max_column

    def _min_column(self, ws_name):
        return self.wb[ws_name]._min_column

    def _has_worksheet(self, ws_name):
        return ws_name in self.wb.sheetnames



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import datetime

import pytest

from efc.utils import col_str_to_index, datetime_to_openxml, parse_date


@pytest.mark.parametrize(
    ['col', 'index'],
    (
            ('A', 1),
            ('C', 3),
            ('Z', 26),
            ('AA', 27),
            ('HPP', 5840),
    )
)
def test_col_str_to_index(col, index):
    assert col_str_to_index(col) == index


@pytest.mark.parametrize(
    ['dt', 'cleaned_dt'],
    (
            ('1', datetime(1900, 1, 1)),
            ('2', datetime(1900, 1, 2)),
            ('61', datetime(1900, 3, 1)),
            ('39448', datetime(2008, 1, 1)),
            ('43831.0234', datetime(2020, 1, 1)),
            ('43832', datetime(2020, 1, 2)),
    )
)
def test_parse_date(dt, cleaned_dt):
    assert parse_date(dt) == cleaned_dt


@pytest.mark.parametrize(
    ['dt', 'openxml_dt'],
    (
            (datetime(1900, 1, 1), '1'),
            (datetime(1900, 1, 2), '2'),
            (datetime(1900, 3, 1), '61'),
            (datetime(2008, 1, 1), '39448'),
            (datetime(2020, 1, 1, minute=33, second=41, microsecond=760000), '43831.0234'),
            (datetime(2020, 1, 2), '43832'),
            (datetime(2020, 1, 1, 1, 0, 1), '43831.041678240741'),
    )
)
def test_datetime_to_openxml(dt, openxml_dt):
    assert datetime_to_openxml(dt) == openxml_dt



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from efc.rpn_builder.lexer import Lexer
from efc.rpn_builder.parser import Parser
from efc.rpn_builder.rpn import RPN

__author__ = 'Gleb Orlov <orlovgb@mail.ru>'
__version__ = '0.2.2'





# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import pytest

from tests.test_rpn.mock import get_calculator


@pytest.fixture(scope='module')
def calculate():
    calculator = get_calculator()
    return lambda line: calculator(line, None, None)


@pytest.mark.parametrize(
    ['line', 'result'],
    (
            ('4', 4),
            ('-4', -4),
            ('4 + 4', 8),
            ('-4 + 4', 0),
            ('4 - 2', 2),
            ('4 * 4', 16),
            ('9 ^ 2', 81),
            ('9 / 3', 3),
            ('1 + 2 * 3', 7),
            ('2 * 3 + 1', 7),
            ('2 - (2 - 3)', 3),
            ('2 - (-2 - 3)', 7),
            ('2 - 2 * 8 - 3', -17),
            ('2 - 2 - 3 - 6', -9),
            ('2 - 2 - 3', -3),
            ('2 - 2 + 3', 3),
            ('2 - (2 + 3)', -3),
    )
)
def test_arithmetic(calculate, line, result):
    assert calculate(line).value == result


@pytest.mark.parametrize(
    ['line', 'result'],
    (
            ('4 = 3', False),
            ('4 <> 3', True),
            ('4 > 3', True),
            ('4 >= 4', True),
            ('4 < 3', False),
            ('3 <= 3', True),
            ('4 + 1 > 4', True),
            ('4 > 4 - 3', True),
            ('4 * 2 + 2 <> 4 / 3 - 1', True),

            ('1 > TRUE', False),
            ('1 > FALSE', False),
            ('1 > "1"', False),
            ('"1" > 1', True),

            ('1 = TRUE', False),
            ('1 = FALSE', False),
            ('1 = "1"', False),
            ('"1" = 1', False),

            ('1 < TRUE', True),
            ('1 < FALSE', True),
            ('1 < "1"', True),
            ('"1" < 1', False),
    )
)
def test_compare(calculate, line, result):
    assert calculate(line).value == result


@pytest.mark.parametrize(
    ['line', 'result'],
    (
            ('"test" & "2"', 'test2'),
            ('2 & 3', '23'),
    )
)
def test_concat(calculate, line, result):
    assert calculate(line).value == result



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from abc import ABCMeta, abstractmethod

from six import add_metaclass


@add_metaclass(ABCMeta)
class BaseCache(object):
    def __init__(self):
        self._items = {}

    def clear(self):
        self._items.clear()

    @abstractmethod
    def remove_cell(self, ws_name, row, column):
        pass

    def __getitem__(self, item):
        return self._items[item]

    def get(self, key, default=None):
        try:
            return self._items[key]
        except KeyError:
            return default

    def __setitem__(self, key, value):
        self._items[key] = value

    def __contains__(self, item):
        return item in self._items


class SingleCellCache(BaseCache):
    def remove_cell(self, ws_name, row, column):
        for row_fixed, column_fixed in ((False, False), (False, True), (True, False), (True, True)):
            key = (ws_name, row, column, row_fixed, column_fixed)
            if key in self._items:
                del self._items[key]


class RangeCache(BaseCache):
    def remove_cell(self, ws_name, row, column):
        for key in list(self._items):
            if ws_name == key[0]:
                if key[1] is None:
                    if key[2] >= column >= key[4]:
                        del self._items[key]
                elif key[2] is None:
                    if key[1] >= row >= key[3]:
                        del self._items[key]
                elif key[1] >= row >= key[3] and key[2] >= column >= key[4]:
                    del self._items[key]


class CacheManager(object):
    CACHE_TYPES = {
        'single': SingleCellCache,  # cache SingleCellOperand instances
        'cells': SingleCellCache,  # cache calculated Cells
        'range': RangeCache,  # cache CellRangeOperand instances
        'ifs': RangeCache,  # cache IFS expressions for ranges
    }

    def __init__(self):
        self._caches = {}

    def __getitem__(self, item):
        if item not in self._caches:
            self._caches[item] = self.CACHE_TYPES[item]()
        return self._caches[item]

    def clear(self):
        self._caches.clear()

    def remove_cell(self, ws_name, row, column):
        for cache in self._caches.values():
            cache.remove_cell(ws_name, row, column)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from efc.base.errors import BaseEFCException


class BaseInterfaceError(BaseEFCException):
    code = 400


class NamedRangeNotFound(BaseInterfaceError):
    code = 401
    msg = 'Named range not found'



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from six import python_2_unicode_compatible

from efc.utils import col_str_to_index

__all__ = ('FloatToken', 'IntToken', 'BoolToken',
           'StringToken', 'FunctionToken', 'CellsRangeToken',
           'SingleCellToken', 'NamedRangeToken', 'AddToken',
           'SubtractToken', 'DivideToken', 'MultiplyToken',
           'ConcatToken', 'ExponentToken', 'CompareNotEqToken',
           'CompareGTEToken', 'CompareLTEToken', 'CompareGTToken',
           'CompareLTToken', 'CompareEqToken', 'LeftBracketToken',
           'RightBracketToken', 'SpaceToken', 'Separator',
           'OperandToken', 'OperationToken', 'ArithmeticToken')


@python_2_unicode_compatible
class Token(object):
    pattern = None

    def __init__(self, match):
        self._match = match
        self.src_value = match.group(0)
        self.token_value = self.get_value(match.groupdict())

    @classmethod
    def get_group_pattern(cls):
        return r'(?P<%s>%s)' % (cls.__name__, cls.pattern)

    def get_value(self, m):
        return m[self.__class__.__name__]

    def __str__(self):
        return '<%s, %s>' % (self.__class__.__name__, self.token_value)


class OperandToken(Token):
    pass


class FloatToken(OperandToken):
    pattern = r'(?P<float_value>\d+\.\d+)((?P<float_percent>%)|\b)'

    def get_value(self, m):
        v = float(m['float_value'])
        if m['float_percent'] is not None:
            v /= 100
        return v


class IntToken(OperandToken):
    pattern = r'(?P<int_value>\d+)((?P<int_percent>%)|\b)'

    def get_value(self, m):
        v = int(m['int_value'])
        if m['int_percent'] is not None:
            v /= 100
        return v


class BoolToken(OperandToken):
    pattern = r'\b(TRUE|FALSE)\b'

    def get_value(self, m):
        return super(BoolToken, self).get_value(m) == 'TRUE'


class StringToken(OperandToken):
    pattern = r'"[^"]*"'

    def get_value(self, m):
        return super(StringToken, self).get_value(m)[1:-1]


class AddressToken(OperandToken):
    @staticmethod
    def clean_ws_name(v):
        if v and v.startswith('\''):
            return v[1:-1]
        return v

    def __getattr__(self, item):
        return self.token_value[item]


class SingleCellToken(AddressToken):
    pattern = (r"((?P<q1>')?(\[(?P<s_doc>\w+)\])?(?P<single_ws_name>(?(q1)[^']|\w)+)?(?(q1)'|)!)?"
               r"(?P<column_fixed>\$)?(?P<column>[A-Z]+)(?P<row_fixed>\$)?(?P<row>[0-9]+)\b")

    def get_value(self, m):
        return {
            'ws_name': self.clean_ws_name(m['single_ws_name']),
            'row': int(m['row']),
            'column': col_str_to_index(m['column']),
            'row_fixed': bool(m['row_fixed']),
            'column_fixed': bool(m['column_fixed']),
        }


class CellsRangeToken(AddressToken):
    pattern = (r"((?P<q2>')?(\[(?P<r_doc>\w+)\])?(?P<range_ws_name>(?(q2)[^']|\w)+)?(?(q2)'|)!)?"
               r"((?P<column1_fixed>\$)?(?P<column1>[A-Z]+))?((?P<row1_fixed>\$)?(?P<row1>[0-9]+))?"
               r":((?P<column2_fixed>\$)?(?P<column2>[A-Z]+))?((?P<row2_fixed>\$)?(?P<row2>[0-9]+))?\b")

    def get_value(self, m):
        return {
            'ws_name': self.clean_ws_name(m['range_ws_name']),
            'row1': int(m['row1']) if m['row1'] is not None else None,
            'column1': col_str_to_index(m['column1']) if m['column1'] is not None else None,
            'row2': int(m['row2']) if m['row2'] is not None else None,
            'column2': col_str_to_index(m['column2']) if m['column2'] is not None else None,
            'row1_fixed': bool(m['row1_fixed']),
            'row2_fixed': bool(m['row2_fixed']),
            'column1_fixed': bool(m['column1_fixed']),
            'column2_fixed': bool(m['column2_fixed']),
        }


class NamedRangeToken(AddressToken):
    pattern = (r"((?P<q3>')?(\[(?P<n_doc>\w+)\])?(?P<named_range_ws_name>(?(q3)[^']|\w)+)?(?(q3)'|)!)?"
               r"(?P<range_name>\w+)")

    def get_value(self, m):
        return {
            'ws_name': self.clean_ws_name(m['named_range_ws_name']),
            'name': m['range_name']
        }


class OperationToken(Token):
    pass


class FunctionToken(OperationToken):
    pattern = r'[A-Z]+(?=\()'


class ArithmeticToken(OperationToken):
    pass


class AddToken(ArithmeticToken):
    pattern = r'\+'


class SubtractToken(ArithmeticToken):
    pattern = r'\-'


class DivideToken(ArithmeticToken):
    pattern = r'/'


class MultiplyToken(ArithmeticToken):
    pattern = r'\*'


class ConcatToken(ArithmeticToken):
    pattern = r'\&'


class ExponentToken(ArithmeticToken):
    pattern = r'\^'


class CompareNotEqToken(ArithmeticToken):
    pattern = r'\<\>'


class CompareGTEToken(ArithmeticToken):
    pattern = r'\>\='


class CompareLTEToken(ArithmeticToken):
    pattern = r'\<\='


class CompareGTToken(ArithmeticToken):
    pattern = r'\>'


class CompareLTToken(ArithmeticToken):
    pattern = r'\<'


class CompareEqToken(ArithmeticToken):
    pattern = r'\='


class LeftBracketToken(Token):
    pattern = r'\('


class RightBracketToken(Token):
    pattern = r'\)'


class SpaceToken(Token):
    pattern = r'[ \n]+'


class Separator(Token):
    pattern = r','



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from itertools import chain

from efc import Lexer, Parser
from efc.interfaces.base import BaseExcelInterface
from efc.rpn_builder.parser.operands import CellSetOperand, SingleCellOperand


class ExcelMock(BaseExcelInterface):
    data = {
        'Sheet 1': {
            1: {1: 13, 2: 16, 3: 18},
            3: {1: 4, 2: 2, 3: 8},
        },
        'Yet another sheet': {
            100: {1: 4, 2: 2, 3: 8},
            104: {1: 4, 2: 2, 3: 8, 27: 45},
        },
        'Sheet4': {
            1: {1: 13, 2: 16, 3: 18},
            2: {1: 13, 2: 16, 3: 18},
            3: {1: 4, 2: 2, 3: 8},
        },
        'Sheet5': {
            1: {1: '', 2: 16, 3: None},
            2: {1: 13, 2: '', 3: 18},
            3: {1: None, 2: 2, 3: ''},
        },
    }

    def _cell_to_value(self, row, column, ws_name):
        return self.data[ws_name].get(row, {}).get(column)

    @property
    def named_ranges(self):
        op_set = CellSetOperand(ws_name='Sheet 1', source=self)
        op_set.add_row([SingleCellOperand(1, 2, ws_name='Sheet 1', source=self),
                        SingleCellOperand(1, 3, ws_name='Sheet 1', source=self)])
        return {
            'test': SingleCellOperand(1, 2, ws_name='Sheet 1', source=self),
            'test2': op_set
        }

    def _get_named_range_formula(self, name, ws_name):
        pass

    def _named_range_to_cells(self, name, ws_name):
        return self.named_ranges[name]

    def _has_worksheet(self, ws_name):
        return ws_name in self.data

    def _has_named_range(self, name, ws_name):
        return name in self.named_ranges

    def _min_row(self, ws_name):
        return min(self.data[ws_name])

    def _min_column(self, ws_name):
        return min(chain(*self.data[ws_name].values()))

    def _max_row(self, ws_name):
        return max(self.data[ws_name])

    def _max_column(self, ws_name):
        return max(chain(*self.data[ws_name].values()))


def get_calculator():
    lexer = Lexer()
    parser = Parser()

    def calculate(formula, ws_name, source):
        tokens_line = lexer.parse(formula)
        rpn = parser.to_rpn(tokens_line, ws_name, source)
        return rpn.calc(ws_name, source)

    return calculate



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from datetime import datetime
from os.path import dirname, join

import pytest
from openpyxl import load_workbook

from efc.interfaces.iopenpyxl import OpenpyxlInterface


@pytest.fixture(scope='function')
def workbook():
    path = join(dirname(__file__), 'fixtures', 'openpyxl.xlsx')
    return load_workbook(path)


@pytest.fixture(scope='function')
def interface(workbook):
    return OpenpyxlInterface(workbook, use_cache=True)


@pytest.mark.parametrize(
    ('cell', 'result'),
    (
            ('A1', 1),
            ('A2', 2),
            ('A3', 3),
            ('A4', 4),
            ('A5', 5),
    )
)
def test_simple_cell_value(cell, result, interface):
    assert interface.calc_cell(cell, 'ws1') == result


@pytest.mark.parametrize(
    ('cell', 'result'),
    (
            ('B1', 2),
            ('B2', 3),
            ('B3', 4),
            ('B4', 5),
            ('B5', 6),
    )
)
def test_simple_formula_value(cell, result, interface):
    assert interface.calc_cell(cell, 'ws1') == result


@pytest.mark.parametrize(
    ('cell', 'result'),
    (
            ('D1', 15),
            ('D2', 20),
            ('D3', 15),
            ('D4', 46),
    )
)
def test_sum_formula_value(cell, result, interface):
    assert interface.calc_cell(cell, 'ws1') == result


@pytest.mark.parametrize(
    ('cell', 'result'),
    (
            ('A7', datetime(2020, 1, 1)),
            ('B7', datetime(2020, 1, 2)),
            ('C7', 1),
    )
)
def test_date_operations_value(cell, result, interface):
    assert interface.calc_cell(cell, 'ws1') == result


@pytest.mark.parametrize(
    ('cell', 'result'),
    (
            ('A8', 'text1'),
            ('B8', 'text2'),
            ('C8', 'text1text2'),
    )
)
def test_test_operations_value(cell, result, interface):
    assert interface.calc_cell(cell, 'ws1') == result


def test_openpyxl_cache_enabled(workbook, interface):
    assert interface.calc_cell('B1', 'ws1') == 2
    workbook['ws1']['A1'].value = 3
    assert interface.calc_cell('B1', 'ws1') == 2  # value from cache
    interface.clear_cache()
    assert interface.calc_cell('B1', 'ws1') == 4


def test_openpyxl_cache_disabled(workbook):
    interface = OpenpyxlInterface(workbook, use_cache=False)
    assert interface.calc_cell('B1', 'ws1') == 2
    workbook['ws1']['A1'].value = 3
    assert interface.calc_cell('B1', 'ws1') == 4



# Copyright (c) 2019 Gleb Orlov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from __future__ import absolute_import, division, print_function, unicode_literals

from setuptools import find_packages, setup

import efc

classifiers = [
    'Development Status :: 2 - Pre-Alpha',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 3',
    'Intended Audience :: Developers',
    'License :: OSI Approved :: MIT License',
]

with open('README.md', 'r') as fp:
    long_description = fp.read()

setup(name='excel-formulas-calculator',
      version=efc.__version__,
      author='Gleb Orlov',
      author_email='orlovgb@mail.ru',
      url='https://github.com/ulalka/excel-formulas-calculator',
      packages=find_packages(),
      include_package_data=True,
      install_requires=['six'],
      extras_require={
          'openpyxl': ['openpyxl'],
      },
      tests_require=['pytest'],
      test_suite='tests',
      license='MIT',
      description='Library to calculate excel formulas',
      long_description=long_description,
      long_description_content_type='text/markdown',
      classifiers=classifiers,
      python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, <4',
      )



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

import pytest

from efc.rpn_builder.parser.operands import BadReference, NumErrorOperand, ValueNotAvailable
from .mock import ExcelMock, get_calculator


@pytest.fixture(scope='session')
def calc():
    source = ExcelMock()
    calculator = get_calculator()
    return lambda line, ws_name: calculator(line, ws_name, source)


def test_SUM(calc):
    assert calc('SUM(Sheet4!A1:B3)', 'Yet another sheet').value == 64
    assert calc('SUM([0]Sheet4!A1:B3)', 'Yet another sheet').value == 64
    assert calc('SUM(Sheet4!A1:B3) + 1', 'Yet another sheet').value == 65
    assert calc('SUM(Sheet4!A1:B3,A2:B3)', 'Sheet4').value == 99
    assert calc('SUM(Sheet4!A1:B3,SUM(A3:B3))', 'Sheet4').value == 70


def test_SUMIFS(calc):
    assert calc('SUMIFS(Sheet4!A1:B3,Sheet4!A1:B3,">4")', 'Yet another sheet').value == 58
    assert calc('SUMIFS(Sheet4!A1:B3,Sheet4!A1:B3,"13")', 'Yet another sheet').value == 26


def test_SUMIF(calc):
    assert calc('SUMIF(Sheet4!A1:B3,">4",Sheet4!A1:B3)', 'Yet another sheet').value == 58
    assert calc('SUMIF(Sheet4!A1:B3,"13",Sheet4!A1:B3)', 'Yet another sheet').value == 26


def test_MOD(calc):
    assert calc('MOD(\'Sheet 1\'!B3,4)', 'Yet another sheet').value == 2
    assert calc('MOD(\'Sheet 1\'!A3,\'Sheet 1\'!C3)', 'Yet another sheet').value == 4
    assert calc('MOD(\'Sheet 1\'!A3,\'Sheet 1\'!B3 * 2)', 'Yet another sheet').value == 0


def test_IF(calc):
    assert calc('IF(2>1,1,2)', 'Yet another sheet').value == 1
    assert calc('IF(TRUE,1,2)', 'Yet another sheet').value == 1
    assert calc('IF(FALSE,1,2)', 'Yet another sheet').value == 2
    assert calc('IF(TRUE,1,2 ** 5)', 'Yet another sheet').value == 1
    assert calc('IF(\'Sheet 1\'!A3 = 4,\'Sheet 1\'!C3, 0)', 'Yet another sheet').value == 8


def test_IFS(calc):
    assert calc('IFS(FALSE,1,FALSE,2,TRUE,3)', 'Yet another sheet').value == 3
    assert calc('IFS(TRUE,1,FALSE,2,TRUE,3)', 'Yet another sheet').value == 1
    assert calc('IFS(FALSE,1,TRUE,2,TRUE,3)', 'Yet another sheet').value == 2

    with pytest.raises(ValueNotAvailable):
        assert calc('IFS(FALSE,1,FALSE,2,FALSE,3)', 'Yet another sheet').value == 1


def test_IFERROR(calc):
    assert calc('IFERROR(5/0,1)', 'Yet another sheet').value == 1
    assert calc('IFERROR(5+6, 0)', 'Yet another sheet').value == 11


def test_MAX(calc):
    assert calc('MAX(Sheet4!A1:B3)', 'Yet another sheet').value == 16
    assert calc('MAX(Sheet4!A1:B3,100)', 'Yet another sheet').value == 100


def test_MIN(calc):
    assert calc('MIN(Sheet4!A1:B3)', 'Yet another sheet').value == 2
    assert calc('MIN(Sheet4!A1:B3,1)', 'Yet another sheet').value == 1


def test_LEFT(calc):
    assert calc('LEFT("test", 2)', 'Yet another sheet').value == 'te'


def test_RIGHT(calc):
    assert calc('RIGHT("test", 2)', 'Yet another sheet').value == 'st'


def test_MID(calc):
    assert calc('MID("hello",2,2)', 'Sheet 1').value == 'el'


def test_ISBLANK(calc):
    assert calc('ISBLANK("test")', 'Yet another sheet').value is False
    assert calc('ISBLANK("")', 'Yet another sheet').value is False
    assert calc('ISBLANK(Sheet4!AA1)', 'Yet another sheet').value is True


def test_OR(calc):
    assert calc('OR(0,0,0,TRUE)', 'Yet another sheet').value is True
    assert calc('OR(FALSE, 0)', 'Yet another sheet').value is False
    assert calc('OR(FALSE, 0 + 2)', 'Yet another sheet').value is True


def test_AND(calc):
    assert calc('AND(1,1,1,TRUE)', 'Yet another sheet').value is True
    assert calc('AND(FALSE, 0)', 'Yet another sheet').value is False
    assert calc('AND(TRUE, 0 + 2)', 'Yet another sheet').value is True


def test_not(calc):
    assert calc('NOT(1)', 'Yet another sheet').value is False
    assert calc('NOT(0)', 'Yet another sheet').value is True
    assert calc('NOT(123)', 'Yet another sheet').value is False
    assert calc('NOT(TRUE)', 'Yet another sheet').value is False
    assert calc('NOT(FALSE)', 'Yet another sheet').value is True
    assert calc('NOT(A1)', 'Yet another sheet').value is True
    assert calc('NOT(A1)', 'Sheet4').value is False
    assert calc('NOT("")', 'Yet another sheet').value is True
    assert calc('NOT(NOT(""))', 'Yet another sheet').value is False


def test_ROUND(calc):
    assert calc('ROUND(2.3456, 1)', 'Yet another sheet').value == 2.3
    assert calc('ROUND(2, 2)', 'Yet another sheet').value == 2.0
    assert calc('ROUND("2.34567", 2)', 'Yet another sheet').value == 2.35


def test_ROUNDDOWN(calc):
    assert calc('ROUNDDOWN(1.345,0)', 'Sheet 1').value == 1.0
    assert calc('ROUNDDOWN(1.345,1)', 'Sheet 1').value == 1.3
    assert calc('ROUNDDOWN(1.345,2)', 'Sheet 1').value == 1.34


def test_FLOOR(calc):
    assert calc('FLOOR(10,3)', 'Sheet 1').value == 9
    assert calc('FLOOR(16,7)', 'Sheet 1').value == 14
    assert calc('FLOOR(26,13)', 'Sheet 1').value == 26


def test_COUNT(calc):
    assert calc('COUNT(1.3456, 1, "test")', 'Yet another sheet').value == 2
    assert calc('COUNT(A1:C4)', 'Sheet 1').value == 6


def test_COUNTIF(calc):
    assert calc('COUNTIF(A1:C4, ">4")', 'Sheet 1').value == 4
    assert calc('COUNTIF(A1:C4, "13")', 'Sheet4').value == 2


def test_COUNTBLANK(calc):
    assert calc('COUNTBLANK(A1:C4)', 'Sheet 1').value == 6
    assert calc('COUNTBLANK(A1:B4)', 'Sheet4').value == 2


def test_ABS(calc):
    assert calc('ABS(1.32)', 'Sheet 1').value == 1.32
    assert calc('ABS(-42)', 'Sheet4').value == 42


def test_OFFSET(calc):
    assert calc('OFFSET(A1,2,1)', 'Sheet 1').value == 2
    assert calc('OFFSET(A1,2,1,1)', 'Sheet 1').value == 2
    assert calc('OFFSET(A1,B3,1)', 'Sheet 1').value == 2
    assert calc('SUM(OFFSET(A1,2,1,1,2))', 'Sheet 1').value == 10


def test_MATCH(calc):
    assert calc('MATCH(13,Sheet4!A1:A3)', 'Yet another sheet').value == 1


def test_AVERAGE(calc):
    assert calc('AVERAGE(Sheet4!A1:B3)', 'Yet another sheet').value == 64 / 6
    assert calc('AVERAGEIFS(Sheet4!A1:B3,Sheet4!A1:B3,"13")', 'Yet another sheet').value == 13


def test_AVERAGEIFS(calc):
    assert calc('AVERAGEIFS(Sheet4!A1:B3,Sheet4!A1:B3,"13")', 'Yet another sheet').value == 13


def test_VLOOKUP(calc):
    assert calc('VLOOKUP(13,Sheet4!A1:B3,2)', 'Yet another sheet').value == 16


def test_SMALL(calc):
    assert calc('SMALL(Sheet4!A1:B3,1)', 'Yet another sheet').value == 2
    assert calc('SMALL(Sheet4!A1:B3,2)', 'Yet another sheet').value == 4
    assert calc('SMALL(Sheet4!A1:B3,4)', 'Yet another sheet').value == 13


def test_LARGE(calc):
    assert calc('LARGE(Sheet4!A1:B3,1)', 'Yet another sheet').value == 16
    assert calc('LARGE(Sheet4!A1:B3,2)', 'Yet another sheet').value == 16
    assert calc('LARGE(Sheet4!A1:B3,4)', 'Yet another sheet').value == 13


def test_COUNTIFS(calc):
    assert calc('COUNTIFS(Sheet4!A1:B3,Sheet4!A1:B3,">4")', 'Yet another sheet').value == 4
    assert calc('COUNTIFS(Sheet4!A1:B3,Sheet4!A1:B3,"13")', 'Yet another sheet').value == 2


def test_COUNTA(calc):
    assert calc('COUNTA(Sheet4!A1:B4)', 'Yet another sheet').value == 6
    assert calc('COUNTA(Sheet5!A1:B4)', 'Yet another sheet').value == 5


def test_CONCATENATE(calc):
    assert calc('CONCATENATE(Sheet4!A1,Sheet4!B3,"13")', 'Yet another sheet').value == '13213'
    assert calc('CONCATENATE("",Sheet4!B3,TRUE)', 'Yet another sheet').value == '2TRUE'


def test_INDEX(calc):
    assert calc('INDEX(Sheet4!A1:A3,1)', 'Yet another sheet').value == 13  # A1
    assert calc('INDEX(Sheet4!A1:B3,1,2)', 'Yet another sheet').value == 16  # B1
    assert calc('INDEX(Sheet4!A1:A3,3)', 'Yet another sheet').value == 4  # A3
    assert calc('INDEX(Sheet4!A2:A3,2)', 'Yet another sheet').value == 4  # A3

    assert calc('INDEX(Sheet4!A:A,3)', 'Yet another sheet').value == 4  # A3
    assert calc('INDEX(Sheet4!A:A,3,1)', 'Yet another sheet').value == 4  # A3
    assert calc('INDEX(Sheet4!3:3,1,1)', 'Yet another sheet').value == 4  # A3
    assert calc('INDEX(Sheet4!3:3,1,2)', 'Yet another sheet').value == 2  # B3

    with pytest.raises(BadReference):
        assert calc('INDEX(Sheet4!A1:A3,100,1)', 'Yet another sheet').value

    with pytest.raises(BadReference):
        assert calc('INDEX(Sheet4!A1:A3,1,100)', 'Yet another sheet').value

    assert calc('INDEX(Sheet4!A1:C3,1,1)', 'Yet another sheet').value == 13
    assert calc('INDEX(Sheet4!A1:C3,1,3)', 'Yet another sheet').value == 18

    with pytest.raises(BadReference):
        assert calc('INDEX(Sheet4!A1:C3,1)', 'Yet another sheet').value

    with pytest.raises(BadReference):
        assert calc('INDEX(Sheet4!A1:C3,100,1)', 'Yet another sheet').value

    with pytest.raises(BadReference):
        assert calc('INDEX(Sheet4!A1:C3,1,100)', 'Yet another sheet').value

    with pytest.raises(BadReference):
        assert calc('INDEX(Sheet4!A1:C3,0,100)', 'Yet another sheet').value

    with pytest.raises(BadReference):
        assert calc('INDEX(Sheet4!A1:C3,1,0)', 'Yet another sheet').value

    with pytest.raises(BadReference):
        assert calc('INDEX(Sheet4!A:A,3,2)', 'Yet another sheet').value

    with pytest.raises(BadReference):
        assert calc('INDEX(Sheet4!1:1,2)', 'Yet another sheet').value


def test_SUBSTITUTE(calc):
    assert calc('SUBSTITUTE("123123123","1","22")', 'Yet another sheet').value == '222322232223'
    assert calc('SUBSTITUTE("123123123","1","22", 2)', 'Yet another sheet').value == '22232223123'
    assert calc('SUBSTITUTE("123123123","1","22", -1)', 'Yet another sheet').value == '222322232223'


def test_TRIM(calc):
    assert calc('TRIM(1)', 'Yet another sheet').value == '1'
    assert calc('TRIM(0)', 'Yet another sheet').value == '0'
    assert calc('TRIM("1")', 'Yet another sheet').value == '1'
    assert calc('TRIM(" 1 ")', 'Yet another sheet').value == '1'
    assert calc('TRIM(" 1 1 ")', 'Yet another sheet').value == '1 1'
    assert calc('TRIM(" 1      1 ")', 'Yet another sheet').value == '1 1'


def test_LEN(calc):
    assert calc('LEN(1)', 'Yet another sheet').value == 1
    assert calc('LEN(0)', 'Yet another sheet').value == 1
    assert calc('LEN("1")', 'Yet another sheet').value == 1
    assert calc('LEN(" 1 ")', 'Yet another sheet').value == 3


def test_YEARFRAC(calc):
    with pytest.raises(NumErrorOperand):
        assert calc('YEARFRAC(1, 2, 5)', 'Yet another sheet').value

    # 30U/360
    assert calc('YEARFRAC(43159, 43160)', 'Yet another sheet').value == 1 / 360
    assert calc('YEARFRAC(43405, 43465, 0)', 'Yet another sheet').value == 60 / 360
    assert calc('YEARFRAC(43889, 43890, 0)', 'Yet another sheet').value == 1 / 360
    assert calc('YEARFRAC(43889, 43891, 0)', 'Yet another sheet').value == 3 / 360

    # Actual/Actual
    assert calc('YEARFRAC(43889, 43890, 1)', 'Yet another sheet').value == 1 / 366
    assert calc('YEARFRAC(43889, 43891, 1)', 'Yet another sheet').value == 2 / 366
    assert calc('YEARFRAC(43523, 43524, 1)', 'Yet another sheet').value == 1 / 365
    assert calc('YEARFRAC(43523, 43525, 1)', 'Yet another sheet').value == 2 / 365

    # Actual/360
    assert calc('YEARFRAC(43889, 43890, 2)', 'Yet another sheet').value == 1 / 360
    assert calc('YEARFRAC(43889, 43891, 2)', 'Yet another sheet').value == 2 / 360
    assert calc('YEARFRAC(43523, 43524, 2)', 'Yet another sheet').value == 1 / 360
    assert calc('YEARFRAC(43523, 43525, 2)', 'Yet another sheet').value == 2 / 360

    # Actual/365
    assert calc('YEARFRAC(43889, 43890, 3)', 'Yet another sheet').value == 1 / 365
    assert calc('YEARFRAC(43889, 43891, 3)', 'Yet another sheet').value == 2 / 365
    assert calc('YEARFRAC(43523, 43524, 3)', 'Yet another sheet').value == 1 / 365
    assert calc('YEARFRAC(43523, 43525, 3)', 'Yet another sheet').value == 2 / 365

    # 30E/360
    assert calc('YEARFRAC(43159, 43160, 4)', 'Yet another sheet').value == 3 / 360
    assert calc('YEARFRAC(43405, 43465, 4)', 'Yet another sheet').value == 59 / 360
    assert calc('YEARFRAC(43889, 43890, 4)', 'Yet another sheet').value == 1 / 360
    assert calc('YEARFRAC(43889, 43891, 4)', 'Yet another sheet').value == 3 / 360



# -*- coding: utf8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from collections import defaultdict

from six import add_metaclass, itervalues, python_2_unicode_compatible, text_type
from six.moves import range

from efc import settings
from efc.base.errors import BaseEFCException
from efc.rpn_builder.parser.metaclasses import MetaCellRangeOperandCache, MetaSingleCellOperandCache
from efc.utils import cached_property, col_index_to_str, digit, u

__all__ = (
    'Operand', 'ErrorOperand', 'ValueErrorOperand', 'WorksheetNotExist',
    'ZeroDivisionErrorOperand', 'SimpleOperand', 'SingleCellOperand',
    'CellSetOperand', 'SimpleSetOperand', 'NamedRangeOperand', 'CellRangeOperand',
    'FunctionNotSupported', 'NotFoundErrorOperand', 'RPNOperand', 'OperandLikeObject', 'OffsetMixin',
    'SetOperand', 'BadReference', 'ValueNotAvailable', 'EmptyOperand', 'NamedRangeNotExist', 'NumErrorOperand',
)


class OperandLikeObject(object):
    def __init__(self, ws_name=None, source=None, *args, **kwargs):
        super(OperandLikeObject, self).__init__(*args, **kwargs)
        self.ws_name = ws_name
        self.source = source


@python_2_unicode_compatible
class Operand(OperandLikeObject):
    value = None

    @cached_property
    def digit(self):
        """Digit type"""
        return digit(self.value)

    @cached_property
    def string(self):
        """String type"""
        value = self.value

        if isinstance(value, bool):
            return text_type(value).upper()
        elif isinstance(value, float):
            if value % 1 == 0:
                return text_type(int(value))
            else:
                return text_type(value).replace('.', settings.FLOAT_DELIMITER)
        elif value is not None:
            return u(value)
        else:
            return ''

    @cached_property
    def is_blank(self):
        return self.value in {None, ''}

    def __int__(self):
        return int(self.value)

    def __float__(self):
        return float(self.value)

    def __str__(self):
        return self.string

    def __trunc__(self):
        return self.__int__()


class EmptyOperand(OperandLikeObject):
    @property
    def value(self):
        return


class ErrorOperand(OperandLikeObject, BaseEFCException):
    code = 300
    msg = 'Unknown error'
    string_value = '#ERROR!'

    def __init__(self, *args, **kwargs):
        self.formula = kwargs.pop('formula', None)
        super(ErrorOperand, self).__init__(*args, **kwargs)

    @property
    def value(self):
        raise self

    @property
    def string(self):
        return self.string_value

    def __getattr__(self, item):
        raise self


class ValueErrorOperand(ErrorOperand):
    code = 301
    msg = 'Cell value error'
    string_value = '#VALUE!'


class WorksheetNotExist(ErrorOperand):
    code = 302
    msg = 'Worksheet does not exist'
    string_value = '#REF!'


class NamedRangeNotExist(ErrorOperand):
    code = 303
    msg = 'Named range "{name}" does not exist'
    string_value = '#NAME?'

    def __init__(self, name, *args, **kwargs):
        super(NamedRangeNotExist, self).__init__(*args, **kwargs)
        self.name = name


class ZeroDivisionErrorOperand(ErrorOperand):
    code = 304
    msg = 'Zero division'
    string_value = '#DIV/0!'


class NotFoundErrorOperand(ErrorOperand):
    code = 305
    msg = 'Result not found'
    string_value = '#VALUE!'


class FunctionNotSupported(ErrorOperand):
    code = 306
    msg = 'Function "{f_name}" not found among available functions'
    string_value = '#NAME?'

    def __init__(self, f_name, *args, **kwargs):
        super(FunctionNotSupported, self).__init__(*args, **kwargs)
        self.f_name = f_name


class BadReference(ErrorOperand):
    code = 307
    msg = 'Bad reference'
    string_value = '#REF!'


class ValueNotAvailable(ErrorOperand):
    code = 308
    msg = 'Value not available'
    string_value = '#N/A'


class NumErrorOperand(ErrorOperand):
    code = 309
    msg = 'Num error'
    string_value = '#NUM!'


class SimpleOperand(Operand):
    def __init__(self, value, *args, **kwargs):
        super(SimpleOperand, self).__init__(*args, **kwargs)
        self.value = value


class CellsOperand(OperandLikeObject):
    def address_to_value(self):
        raise NotImplementedError

    @cached_property
    def value(self):
        if self.source._has_worksheet(self.ws_name):
            return self.address_to_value()
        else:
            raise WorksheetNotExist(ws_name=self.ws_name)

    def get_iter(self):
        raise NotImplementedError

    @cached_property
    def cached_iterable_items(self):
        return list(self.get_iter())

    def __iter__(self):
        return iter(self.cached_iterable_items)


class OffsetMixin(object):
    def offset(self, row_offset=0, col_offset=0):
        raise NotImplementedError


@add_metaclass(MetaSingleCellOperandCache)
class SingleCellOperand(CellsOperand, Operand, OffsetMixin):
    def __init__(self, row, column, row_fixed=False, column_fixed=False, *args, **kwargs):
        super(SingleCellOperand, self).__init__(*args, **kwargs)
        self.row = row
        self.column = column
        self.row_fixed = row_fixed
        self.column_fixed = column_fixed

    def address_to_value(self):
        return self.source._cell_to_value(self.row, self.column, self.ws_name)

    def get_iter(self):
        yield self

    @property
    def address(self):
        return "'%s'!%s%d" % (self.ws_name, col_index_to_str(self.column), self.row)

    def offset(self, row_offset=0, col_offset=0):
        row = self.row
        column = self.column

        if not self.row_fixed:
            row += row_offset
        if not self.column_fixed:
            column += col_offset

        return SingleCellOperand(row=row, column=column,
                                 row_fixed=self.row_fixed, column_fixed=self.column_fixed,
                                 ws_name=self.ws_name, source=self.source)


class SetOperand(OperandLikeObject):
    operands_type = None

    def __init__(self, *args, **kwargs):
        super(SetOperand, self).__init__(*args, **kwargs)
        self._items = defaultdict(list)

    def check_type(self, items):
        if isinstance(items, list):
            if any(not isinstance(i, self.operands_type) for i in items):
                raise ValueErrorOperand()
        elif not isinstance(items, (self.operands_type, ValueNotAvailable)):
            raise ValueErrorOperand()

    def add_cell(self, item, row=0):
        self.check_type(item)
        self._items[row].append(item)

    def add_many(self, items, row=0):
        self.check_type(items)
        append = self._items[row].append
        for item in items:
            append(item)

    def add_row(self, items):
        self.check_type(items)
        r = max(self._items) + 1 if self._items else 1
        self.add_many(items, r)

    def get_iter(self):
        for r in sorted(self._items):
            for item in self._items[r]:
                yield item

    def __iter__(self):
        return self.get_iter()

    @cached_property
    def value(self):
        return list(self)

    def get_cell(self, row, column):
        try:
            return self._items[row - 1][column - 1]
        except KeyError:
            return BadReference()

    @property
    def rows_count(self):
        return len(self._items)

    @property
    def columns_count(self):
        return max(len(c) for c in itervalues(self._items))


class CellSetOperand(SetOperand):
    operands_type = SingleCellOperand


class SimpleSetOperand(SetOperand):
    operands_type = SimpleOperand


@add_metaclass(MetaCellRangeOperandCache)
class CellRangeOperand(CellsOperand, OffsetMixin):
    def __init__(self, row1, column1, row2, column2,
                 row1_fixed=False, column1_fixed=False, row2_fixed=False, column2_fixed=False,
                 *args, **kwargs):
        super(CellRangeOperand, self).__init__(*args, **kwargs)
        self.row1 = row1
        self.column1 = column1
        self.row2 = row2
        self.column2 = column2

        self.row1_fixed = row1_fixed
        self.column1_fixed = column1_fixed
        self.row2_fixed = row2_fixed
        self.column2_fixed = column2_fixed

    def get_iter(self):
        column1 = self.source._min_column(self.ws_name) if self.column1 is None else self.column1
        column2 = self.source._max_column(self.ws_name) if self.column2 is None else self.column2
        row1 = self.source._min_row(self.ws_name) if self.row1 is None else self.row1
        row2 = self.source._max_row(self.ws_name) if self.row2 is None else self.row2

        if row1 == row2:
            for c in range(column1, column2 + 1):
                yield SingleCellOperand(row1, c, ws_name=self.ws_name, source=self.source)
        elif column1 == column2:
            for r in range(row1, row2 + 1):
                yield SingleCellOperand(r, column1, ws_name=self.ws_name, source=self.source)
        else:
            for r in range(row1, row2 + 1):
                for c in range(column1, column2 + 1):
                    yield SingleCellOperand(r, c, ws_name=self.ws_name, source=self.source)

    def address_to_value(self):
        return self.cached_iterable_items

    @property
    def address(self):
        if self.column1 is not None and self.row1 is not None:
            return "'%s'!%s%d:%s%d" % (self.ws_name,
                                       col_index_to_str(self.column1), self.row1,
                                       col_index_to_str(self.column2), self.row2)
        elif self.column1 is not None:
            return "'%s'!%s:%s" % (self.ws_name,
                                   col_index_to_str(self.column1),
                                   col_index_to_str(self.column2))
        else:
            return "'%s'!%d:%d" % (self.ws_name, self.row1, self.row2)

    def offset(self, row_offset=0, col_offset=0):
        row1 = self.row1
        column1 = self.column1
        row2 = self.row2
        column2 = self.column2

        if not self.row1_fixed and self.row1 is not None:
            row1 += row_offset
        if not self.column1_fixed and self.column1 is not None:
            column1 += col_offset
        if not self.row2_fixed and self.row2 is not None:
            row2 += row_offset
        if not self.column2_fixed and self.column2 is not None:
            column2 += col_offset

        return CellRangeOperand(row1=row1, column1=column1,
                                row2=row2, column2=column2,
                                row1_fixed=self.row1_fixed, column1_fixed=self.column1_fixed,
                                row2_fixed=self.row2_fixed, column2_fixed=self.column2_fixed,
                                ws_name=self.ws_name, source=self.source)

    def get_cell(self, row, column):
        row = (self.row1 or 1) + row - 1
        column = (self.column1 or 1) + column - 1

        if self.row1 is None or self.row1 <= row <= self.row2:
            if self.column1 is None or self.column1 <= column <= self.column2:
                return SingleCellOperand(row, column, ws_name=self.ws_name, source=self.source)
        return BadReference()


class NamedRangeOperand(CellsOperand):
    def __init__(self, name, *args, **kwargs):
        super(NamedRangeOperand, self).__init__(*args, **kwargs)
        self.name = name

    def address_to_value(self):
        return self.source._named_range_to_cells(self.name, self.ws_name)

    @cached_property
    def value(self):
        if self.ws_name and not self.source._has_worksheet(self.ws_name):
            raise WorksheetNotExist(ws_name=self.ws_name)
        elif not self.source._has_named_range(self.name, self.ws_name):
            raise NamedRangeNotExist(self.name, self.ws_name)
        else:
            return self.address_to_value()

    def get_iter(self):
        return iter(self.value)


@python_2_unicode_compatible
class RPNOperand(OperandLikeObject, OffsetMixin):
    def __init__(self, rpn, *args, **kwargs):
        super(RPNOperand, self).__init__(*args, **kwargs)
        self.rpn = rpn
        self._result = None

    @cached_property
    def evaluated_value(self):
        v = self.rpn.calc(ws_name=self.ws_name, source=self.source)
        if isinstance(v, RPNOperand):
            v = v.evaluated_value
        return v

    def __getattr__(self, item):
        return getattr(self.evaluated_value, item)

    def offset(self, row_offset=0, col_offset=0):
        return RPNOperand(rpn=self.rpn.offset(row_offset, col_offset), ws_name=self.ws_name, source=self.source)

    def __int__(self):
        return int(self.evaluated_value)

    def __float__(self):
        return float(self.evaluated_value)

    def __str__(self):
        return text_type(self.evaluated_value)

    def __trunc__(self):
        return self.__int__()

    def __iter__(self):
        return iter(self.evaluated_value)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from six.moves import range

from efc.rpn_builder.errors import OperandsMissing
from efc.rpn_builder.parser.operands import (CellSetOperand, ErrorOperand, OffsetMixin, OperandLikeObject,
                                             SimpleOperand, SimpleSetOperand, SingleCellOperand, ValueErrorOperand,
                                             ZeroDivisionErrorOperand, FunctionNotSupported)
from efc.rpn_builder.parser.operations import Operation
from efc.utils import Array

__all__ = ('RPN',)


class RPN(Array):
    def __init__(self, formula):
        super(RPN, self).__init__()
        self.formula = formula

    @staticmethod
    def handle_result(result, ws_name, source):
        if len(result) == 1:
            return result[0]
        else:
            # Trying to get first error in result
            for item in (i for i in result if isinstance(i, ErrorOperand)):
                return item

            if isinstance(result[0], SingleCellOperand):
                set_type = CellSetOperand
            elif isinstance(result[0], SimpleOperand):
                set_type = SimpleSetOperand
            else:
                return result

            # Trying to build set from result
            try:
                result_set = set_type(ws_name=ws_name, source=source)
                result_set.add_row(result)
                return result_set
            except ValueErrorOperand as err:
                return err

    def calc(self, ws_name, source):
        self.reset()
        result = []

        result_append = result.append
        result_pop = result.pop
        for token in self:
            if isinstance(token, OperandLikeObject):
                result_append(token)
            elif isinstance(token, Operation):
                try:
                    args = [result_pop() for _ in range(token.operands_count)]
                except IndexError:
                    raise OperandsMissing(token.f_name, self.formula)

                args.reverse()

                try:
                    v = token.eval(*args)
                except FunctionNotSupported:
                    raise
                except ErrorOperand as err:
                    v = err
                except (TypeError, ValueError):
                    v = ValueErrorOperand()
                except ZeroDivisionError:
                    v = ZeroDivisionErrorOperand()

                if isinstance(v, ErrorOperand):
                    v.formula = self.formula
                    if v.ws_name is None:
                        v.ws_name = ws_name
                    if isinstance(v, FunctionNotSupported):
                        raise v

                result_append(v)

        return self.handle_result(result, ws_name, source)

    def offset(self, row_offset=0, col_offset=0):
        self.reset()

        new_rpn = RPN(formula=self.formula)
        for token in self:
            if isinstance(token, OffsetMixin):
                new_token = token.offset(row_offset=row_offset, col_offset=col_offset)
            else:
                new_token = token
            new_rpn.append(new_token)
        return new_rpn



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from efc.base.errors import BaseEFCException


class RPNError(BaseEFCException):
    pass


class OperandsMissing(RPNError):
    code = 300
    msg = 'The number of operands is more than available in stack for function "{f_name}"'

    def __init__(self, f_name, formula):
        self.f_name = f_name
        self.formula = formula



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals

from efc.base.errors import BaseEFCException


class LexerError(BaseEFCException):
    pass


class CheckSumError(LexerError):
    code = 100
    msg = 'Some symbols from line are lost. Src line: {src_line}. Parsed line: {parsed_line}.'

    def __init__(self, src_line, parsed_line):
        self.src_line = src_line
        self.parsed_line = parsed_line
